/*
 ****************************************************************
 * Licensed Materials - Property of IBM
 * 5725-F96 IBM MessageSight
 * (C) Copyright IBM Corp. 2014.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 ****************************************************************
 */

package com.ibm.ima.samples.plugin.restmsg;

import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import com.ibm.ima.plugin.ImaConnection;
import com.ibm.ima.plugin.ImaConnectionListener;
import com.ibm.ima.plugin.ImaDestinationType;
import com.ibm.ima.plugin.ImaMessage;
import com.ibm.ima.plugin.ImaMessageType;
import com.ibm.ima.plugin.ImaPlugin;
import com.ibm.ima.plugin.ImaReliability;
import com.ibm.ima.plugin.ImaSubscription;
import com.ibm.ima.plugin.ImaSubscriptionType;
import com.ibm.ima.plugin.util.ImaJson;

/**
 * The RestMsgConnection class is the connection listener for the restmsg plug-in.
 * This object represents the connection from the plug-in point of view.
 * <p>
 * The basic job of this listener is to convert data between the external REST 
 * requests and the MessageSight plug-in methods.  The restmsg sample 
 * protocol is conceptually very similar to the natively supported messaging 
 * protocols.  So this example demonstrates the mechanisms for a protocol 
 * plug-in but it does relatively simple conversions from one representation to 
 * another.
 */
public class RestMsgConnection implements ImaConnectionListener {
    ImaConnection  connect;          /* The base connection object this listener is associated with */
    ImaPlugin      plugin;           /* The plug-in object this listener is associated with         */
    Map<String, ImaSubscription> subscriptions = new HashMap<String, ImaSubscription>();
    LinkedBlockingQueue<ImaMessage>receivedMessages = new LinkedBlockingQueue<ImaMessage>();
    static Charset utf8 = Charset.forName("UTF-8");  /* UTF-8 encoding  */
    
    String op = null;
    String path = null;
    String query = null;
    String content_type = null;
    byte[] data = null;
    
    /**
     * Constructor for the connection listener object.
     * 
     * This is called when a connection is created and it saves the connection and plugin objects
     * for later use. 
     */
    RestMsgConnection(ImaConnection connect, ImaPlugin plugin) {
        this.connect = connect;
        this.plugin  = plugin;
    }

    /**
     * Handle the close of a connection.
     * 
     * Nothing to do here. We cannot communicate with
     * the REST client unless it is has made a specific
     * request. The origin of the onClose callback is unknown. 
     * 
     * @see com.ibm.ima.plugin.ImaConnectionListener#onClose(int, java.lang.String)
     */
    public void onClose(int rc, String reason) {
       
    }
    
    /**
     * Handle a request to check if the connection is alive.
     * 
     *  
     * @see com.ibm.ima.plugin.ImaConnectionListener#onLivenessCheck()
     */
    public boolean onLivenessCheck() {
        return true;
    }
    
    /**
     * Parse the query string of a request into key value pairs. 
     * 
     * @param query  The query string from the REST request. 
     * @return		 A Map backed by a HashMap of key value pairs. The
     * 				 key value pairs represent the query string passed in.
     */
    private Map<String, String> parseQuery(String query) {
        if (query == null || query.length() == 0)
            return null;
    	HashMap<String, String> queryMap = new HashMap<String, String>();
    	
    	if (query != null && query.length() > 0) {
    		String[] qArray = query.split("&");    /* TODO: documented as comma, escape not implemented */
    		if (qArray != null && qArray.length > 0) {
    			for (int i=0; i<qArray.length; i++) {
    				String[] keyValue = qArray[i].split("=");
    				if (keyValue != null && keyValue.length == 2) {
    					queryMap.put(keyValue[0], keyValue[1]);
    				}
    			}
    		}
    	}
    	
    	return queryMap;
    }
    
    /**
     * Look for a timeout key within a Map of key value pairs. If
     * the timeout key is present attempt to parse an integer value
     * from it. 
     * 
     * @param queryMap  A Map of key value pairs that represent the query
     * @return          An integer value for timeout if the timeout key exists
     */
    private int getTimeout(Map<String, String> queryMap, String key) {
        if (queryMap == null || key == null)
            return 0;
		String timeoutString = queryMap.get(key);
		int timeout = 0;
		if (timeoutString != null) {
			try {
				timeout = Integer.parseInt(timeoutString);
			} catch (NumberFormatException nfe) {
			}
		}
		return timeout;
    }
    
    /**
     * Handle data from the client.
     * 
     * When we receive data from the client determine the appropriate action to take.
     * If we find one call the method associated with the action.
     * 
     * @see com.ibm.ima.plugin.ImaConnectionListener#onHttpData(java.lang.String, java.lang.String, java.lang.String, java.lang.String, byte[])
     */
    public void onHttpData(String op, String path, String query, String content_type, byte [] data) {
    	
        if (plugin.isTraceable(9)) {
        	plugin.trace("onHttpData op " + op);
        	plugin.trace("onHttpData path " + path);
        	plugin.trace("onHttpData query " + query);
        	plugin.trace("onHttpData content_type " + content_type);
        }
        
        if (connect.getState() == ImaConnection.State_Accepted) { 
        	this.op = op;
        	this.path = path;
        	this.query = query;
        	this.content_type = content_type; 
        	this.data = data;
            connect.setIdentity(connect.getHttpHeader("Client-ID"), connect.getUser(), connect.getPassword());
            return;
        }
        
        if (path.length() < 2 || path.charAt(0) != '/')
            connect.sendHttpData(400, "text/plain;charset=utf-8", "The path is not valid");

    	if ("GET".equals(op)) {
			onHttpDataGET(path, query, content_type, data);
    	} else if ("POST".equals(op)) {
			onHttpDataPOST(path, query, content_type, data);
    	} else if ("PUT".equals(op)) {
    		onHttpDataPUT(path,content_type,data);
    	} else if ("DELETE".equals(op)) {
			connect.sendHttpData(501, "text/plain;charset=utf-8", "");
    	} else {
			connect.sendHttpData(400, "text/plain;charset=utf-8", "The operation is not known");
		}

    }

    /**
     * 
     * Handle a REST GET request. Currently this handles one request - 
     * obtaining any new messages. The method can block if a timeout
     * greater than zero is specified in the query. If no timeout or
     * a timeout equal to zero or less the method will return immediately.
     * 
     * @param path           The REST request path minus leading alias 
     *                       of /restmsg
     * @param query          The query string from the REST request
     * @param content_type   Content type of bytes data
     * @param data           byte array of data associated with request.
     */
    private void onHttpDataGET(String path, String query, String content_type, byte[] data) {
    	ImaMessage msg;

    	/*
    	 * if we have the alias followed by /topic/ for example
    	 *  /restmsg/topic/abc  we want to get retained message 
    	 *  from the topic abc.
    	 */
    	if (path.matches("[/].*\\/topic\\/.+")) {

    		int index = path.indexOf("/topic/") + 7;
    		String topic = path.substring(index);
            connect.getRetainedMessage(topic, 0);
        	connect.sendHttpData(200, "text/plain;charset=utf-8", "Success");
        	return;

    	} else {
    		
    		Map<String, String> queryMap = parseQuery(query);
    		int timeout = getTimeout(queryMap, "timeout");
    		
    		while (true) {  
    			try {
    				if (receivedMessages.isEmpty()){
    					if (timeout <= 0) {
    						timeout = 0;              }
    					msg  = receivedMessages.poll(timeout, TimeUnit.MILLISECONDS);
    				} else {
    					msg  = receivedMessages.poll();
    				}
    				break;
    			} catch (InterruptedException e) {
    				continue;
    			}
    		}
    		
    		// no message to send back
    		if (msg == null) {
    			connect.sendHttpData(204, (String)null, (byte [])null);
    			return;
    		}
    		
    		String subname = "" + msg.getUserData();
    		
    		if (subname != null) {
    			connect.setHttpHeader("Subscription", subname);
    		}
    		
    		if (msg.getDestinationType() == ImaDestinationType.queue) {
    			connect.setHttpHeader("Queue", msg.getDestination());
    		} else {
    			String retainedValue = Boolean.toString(msg.getRetain());
    			connect.setHttpHeader("Retained", retainedValue);
    			connect.setHttpHeader("Topic", msg.getDestination());
    		}
    		
    		ImaMessageType mtype = msg.getMessageType();
    		switch (mtype) {
    			case jmsBytesMessage:
    			case bytes:             
    				content_type = "application/octet-stream";
    				break;
    			case text:
    			case jmsTextMessage:    
    				content_type = "text/plain;charset=utf8";
    				break;
    			case jsonObject:
    			case jsonArray:
    				content_type = "application/json";
    				break;
    			case jmsMessage:
    			case jmsMapMessage:     /* TODO: convert to JSON object */   
    			case jmsObjectMessage:  
    			case jmsStreamMessage:  /* TODO: convert to JSON array */
    			default:
    				content_type = "application/octet-steam";
    				break;
    		}

    		if (plugin.isTraceable(9)) {
    			plugin.trace(9, "onMessage: " + msg);
    		}
    		
    		connect.sendHttpData(200, content_type, msg.getBodyBytes());
    	}

    }
    
    /**
     * Handle a REST PUT request. Currently this handles two requests - 
     * sending a retained message or removing a retained message from
     * a topic.
     * 
     * if data is null or zero length an attempt to remove a retained
     * message from the topic will be performed.
     * 
     * @param path           The REST request path minus leading alias 
     *                       of /restmsg
     * @param content_type   Content type of bytes data
     * @param data           byte array of data associated with request.
     */
    private void onHttpDataPUT(String path,  String content_type, byte[] data) {
    	
    	if (path.matches("[/].*\\/topic\\/.+")) {
    		
    		int index = path.indexOf("/topic/") + 7;
    		String topic = path.substring(index);
    		if (data == null || data.length == 0) {
    			connect.deleteRetained(topic);
            	connect.sendHttpData(200, "text/plain;charset=utf-8", "Success");
    		} else {
    			doSend(topic, false, true, content_type, data);
    		}
    		
    	} else {
    		connect.sendHttpData(400, "text/plain;charset=utf-8", "The path is not valid");
    	}
    	
    }
    
    
    /**
     * Handle a REST POST request. Currently this handles two requests - 
     * subscribing and sending messages. At this point only topics are
     * supported.
     * 
     * @param path           The REST request path minus leading alias 
     *                       of /restmsg
     * @param query          The query string from the REST request
     * @param content_type   Content type of bytes data
     * @param data           byte array of data associated with request.
     */
    private void onHttpDataPOST(String path, String query, String content_type, byte[] data) {
    	Map<String, String> queryMap = parseQuery(query);
    	
    	if (queryMap == null) {
            connect.sendHttpData(400, "text/plain;charset=utf-8", "The query is not valid");
            return;
    	}
    	
    	if (!path.matches("[/].*\\/topic\\/.+")) {
   
    		// check if there is a close query parameter
    		if (queryMap.containsKey("close")) {
    			connect.sendHttpData(200, "text/plain;charset=utf-8", "Success");
    			connect.close();
    		} else {
    			// invalid path - send bad request status
    			connect.sendHttpData(400, "text/plain;charset=utf-8", "The path is not valid");
    		}
    		
    		return;
    	}
    	
		int index = path.indexOf("/topic/") + 7;
    	String subscriptionName = queryMap.get("subscribe");
    	if (subscriptionName != null) {
    		boolean durable = Boolean.parseBoolean(queryMap.get("durable"));
    		String filter = path.substring(index);
    		doSubscribe(filter, subscriptionName, durable, content_type, data);
    	} else {
    		boolean persist = Boolean.parseBoolean(queryMap.get("persist"));
    		boolean retain = Boolean.parseBoolean(queryMap.get("retain"));
    		String topic = path.substring(index);
    		doSend(topic, persist, retain, content_type, data);
    	}
    }
    
    /**
     * Handle data from the client.
     * 
     * This should never be called.
     * 
     * @see com.ibm.ima.plugin.ImaConnectionListener#onData(byte[])
     */
    public int onData(byte[] data, int offset, int length) {
        return length;
    }
    
    /**
     * Handle the completion of a connection and send a connection message to the client.
     * 
     * A non-zero return code indicates that the connection is closed, and that is handled
     * by calling close.
     * 
     * We do not need to do anything as the http response will indicate the connection
     * status.
     * 
     * @see com.ibm.ima.plugin.ImaConnectionListener#onConnected(int, java.lang.String)
     */
    public void onConnected(int rc, String reason) {

    	onHttpData(this.op, this.path, this.query, this.content_type, this.data);
    	
    }
  
    /**
     * Process a publish or send.
     * 
     * Create a MessgeSight message and sent it on the specified destination.
     * Only topic is working for now.
     
     * @param topic				The topic to send a message to
     * @param persist			Specifies if the message should be persisted
     * @param retain            Specifies if the message is of type retained
     * @param content_type		The content type of the bytes data associated
     * 							with the sending of a messages
     * @param data				Byte array of data
     */
    void doSend(String topic, boolean persist, boolean retain, String content_type, byte[] data) {
    	/*
    	 * Check for known text content-types, and otherwise make it a binary
    	 */
    	ImaMessageType mtype = ImaMessageType.bytes;
        if (content_type != null) {
        	if (content_type.startsWith("text/")) {
        	    mtype = ImaMessageType.text;
        	} else if (content_type.startsWith("application/json")) {
        	    mtype = ImaMessageType.jsonObject;
        	} else if (content_type.startsWith("application/xml")) {
                mtype = ImaMessageType.text;
        	}
        }
        
        if (plugin.isTraceable(9)) {
            plugin.trace("doSend Connect=" + this + " Dest=" + topic);
            if (mtype != ImaMessageType.bytes) {
            	plugin.trace("doSend textData=" + new String(data, utf8));
            }
        }
        
        /* Create a message with the appropriate message type.  */
        ImaMessage msg = plugin.createMessage(mtype);
     	msg.setBodyBytes(data);
        
        /* Set attributes and send the message  */
        msg.setRetain(retain);
        msg.setPersistent(persist);

        connect.send(msg, ImaDestinationType.topic, topic, 0);
    	connect.sendHttpData(200, "text/plain;charset=utf-8", "Success");
    }
   
    
    /**
     * Process a subscribe from the client. 
     * 
     * @param filter        	The filter for the subscription.
     * @param name				The name of the subscription.
     * @param durable			Specifies if the subscription is durable or not
     * @param content_type		The content type of the bytes data associated
     * 							with the subscription
     * @param data				Byte array of data
     */
    void doSubscribe(String filter, String name, boolean durable, String content_type, byte[] data) {
    	boolean newSub = false;
    	int id = 0;
        ImaDestinationType desttype = ImaDestinationType.topic;
        
        if (filter == null) {
        	connect.sendHttpData(400, "text/plain;charset=utf-8", "The subscription filter is not valid");
        	return;
        }
        
        ImaSubscriptionType subType = ImaSubscriptionType.NonDurable;
        if (durable) {
            subType = ImaSubscriptionType.Durable;
        }
        
        if (plugin.isTraceable(7)) {
            plugin.trace("doSubscribe Connect=" + this + " Name=" + name + "Dest=" + filter);        
        }
        
        ImaSubscription sub = subscriptions.get(name);
        if (null == sub) {
        	newSub = true;
            sub = connect.newSubscription(desttype, filter, name, subType, false, false);
         	subscriptions.put(name, sub);  
        } else if (desttype != sub.getDestinationType()
        		|| !filter.equals(sub.getDestination())
        		|| subType != sub.getSubscriptionType()) {
        	connect.sendHttpData(400, "text/plain;charset=utf-8", "The subscription request is not valid");
        	return;
        }
        if (!newSub && sub.isSubscribed()) {
        	connect.sendHttpData(400, "text/plain;charset=utf-8", "The subscription already exists");
        	return;
        }
        
        sub.setUserData(id);
        sub.setReliability(ImaReliability.qos1);
        sub.subscribeNoAck(); 
        connect.sendHttpData(200, "text/plain;charset=utf-8", "Success");
    }
    
    /*
     * Process a close subscription.
     * TODO:
     */
    void doCloseSub(ImaJson jobj) {
    	int rc = 0;
    	String reason = null;
    	String name = jobj.getString("Name"); 
    	ImaSubscription subscription = subscriptions.get(name);
    	if (null == subscription) {
    		plugin.trace(5, "doCloseSub: Cannot find subscription for Name='"+name+"' Connect="+this);
    		reason = "doCloseSub: Cannot find subscription for Name='"+name+"'";
    		rc = -3;
    	} else {
            plugin.trace(7, "doCloseSub: Name="+name+" Connect=" + this);
    		subscription.close();
    		if (ImaSubscriptionType.NonDurable == subscription.getSubscriptionType()) {
    			subscriptions.remove(name);
    		}
    	}

    }
    
    /*
     * Process a destroy subscription.
     */
    void doDestroySub(ImaJson jobj) {
    	String reason = "";
    	String name = jobj.getString("Name"); 
    	ImaSubscription subscription = subscriptions.get(name);
    	if (null == subscription) {
    		plugin.trace(5, "doDestroySub: Cannot find subscription: Name=\""+name+"\" Connect=\"" + this + "\"");
    		reason = "Cannot find subscription: Name=\"" + name + "\"";
    		connect.sendHttpData(404, "text/plain;charset=utf-8", reason);
    	} else {
        	plugin.trace(7, "doDestroySub: Name="+name+" Connect=" + this);
    		connect.destroySubscription(null, name, subscription.getSubscriptionType());
   			subscriptions.remove(name);
    	}
    }
    
    
    /**
     * Handle a message from the server.
     * 
     * Enqueue the message for a later get.
     * 
     * @see com.ibm.ima.plugin.ImaConnectionListener#onMessage(com.ibm.ima.plugin.ImaSubscription, com.ibm.ima.plugin.ImaMessage)
     */
    public void onMessage(ImaSubscription sub, ImaMessage msg) {
        msg.setUserData(sub.getName());
        receivedMessages.add(msg);
        /* TODO: flow control */
    }

    /**
     * Handle a completion event.
     * 
     * The completion can come in on a message or subscription.
     * 
     * @see com.ibm.ima.plugin.ImaConnectionListener#onComplete(java.lang.Object, int, java.lang.String)
     */
    public void onComplete(Object obj, int rc, String message) {
        if (plugin.isTraceable(9)) {
            plugin.trace("onComplete Connect=" + this + " Object= " + obj + " rc=" + rc);
        }
        if (obj != null) {
            if (obj instanceof Integer) {
            	if (null != obj) {
            		Integer ID = (Integer)obj;
         
            	}
                /* For now we ignore message completion as only qos=0 (non-acknowledged) publish is supported */
            } else if (obj instanceof ImaSubscription) {
            	ImaSubscription sub = (ImaSubscription)obj;
            	Integer ID =(Integer)sub.getUserData();
            	if (null != ID) {
    
            		sub.setUserData(null);
            	}
            } else {
                plugin.trace(3, "onComplete unknown complete object: Connect=" + this + " Object=" + obj);
            }
        } else {
            plugin.trace(3, "onComplete null object: Connect=" + this);
        }
    }
    
    /**
     * Give a simple string representation of this connection listener object.
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        return "RestMsgConnection " + connect.getClientID();
    }
    
    /**
     * Handle a message from the server.
     * 
     * Queue up message so it may be retrieved by subsequent 
     * calls to get messages.
     * 
     * @see com.ibm.ima.plugin.ImaConnectionListener#onMessage(com.ibm.ima.plugin.ImaSubscription, com.ibm.ima.plugin.ImaMessage)
     */
	public void onGetMessage(Object correlate, int rc, ImaMessage msg) {
		if(rc==0){
	        receivedMessages.add(msg);
		}
	}


}
